/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * hostserv.c
 * Copyright (C) 2003 demond & Amour
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <regex.h>

#include "defs.h"
#include "alloc.h"
#include "channel.h"
#include "chanserv.h"
#include "client.h"
#include "config.h"
#include "hash.h"
#include "helpserv.h"
#include "log.h"
#include "match.h"
#include "memoserv.h"
#include "misc.h"
#include "mystring.h"
#include "nickserv.h"
#include "hostserv.h"
#include "operserv.h"
#include "settings.h"
#include "sprintf_irc.h"
#include "server.h"
#include "timestr.h"
#include "err.h"
#include "hybdefs.h"
#include "sock.h"

#ifdef HAVE_PTHREADS
  #include <pthread.h>
#endif

#ifdef HOSTSERVICES

static void v_help(struct Luser *, int, char **);
static void v_spoof(struct Luser *, int, char **);
static void v_auth(struct Luser *, int, char **);

static void v_access(struct Luser *, int, char **);
static void v_access_add(struct Luser *, struct NickInfo *, int, char **);
static void v_access_del(struct Luser *, struct NickInfo *, int, char **);
static void v_access_list(struct Luser *, struct NickInfo *, int, char **);

static void v_shit(struct Luser *, int, char **);
static void v_shit_add(struct Luser *, int, char **);
static void v_shit_del(struct Luser *, int, char **);
static void v_shit_list(struct Luser *, int, char **);
static void v_shit_save(struct Luser *, int, char **);

#ifdef HAVE_PTHREADS
static void whois_lookup(void *args);
#endif

static struct Command vhostcmds[] = {
  { "HELP", v_help, LVL_NONE },
  { "ACCESS", v_access, LVL_IDENT },
  { "SPOOF", v_spoof, LVL_IDENT },
  { "VHOST", v_spoof, LVL_IDENT },
  { "AUTH", v_auth, LVL_IDENT },
  { "SHITLIST", v_shit, LVL_ADMIN },
  { 0, 0, 0 }
};

static struct Command accesscmds[] = {
  { "ADD", v_access_add, LVL_NONE },
  { "DEL", v_access_del, LVL_NONE },
  { "LIST", v_access_list, LVL_NONE },
  { 0, 0, 0 }
};

static struct Command shitcmds[] = {
  { "ADD", v_shit_add, LVL_ADMIN },
  { "DEL", v_shit_del, LVL_ADMIN },
  { "LIST", v_shit_list, LVL_ADMIN },
  { "SAVE", v_shit_save, LVL_ADMIN },
  { 0, 0, 0 }
};

struct UsedVHost *usedvh;
struct ShitList *shitlist;

/*
vs_process()
  Process command coming from 'nick' directed towards n_HostServ
*/

void
vs_process(char *nick, char *command)

{
  int acnt;
  char **arv;
  struct Command *cptr;
  struct Luser *lptr;
  struct NickInfo *nptr;

  if (!command || !(lptr = FindClient(nick)))
    return;

  if (Network->flags & NET_OFF)
  {
    notice(n_HostServ, lptr->nick,
      "Services are currently \002disabled\002");
    return;
  }

  acnt = SplitBuf(command, &arv);
  if (acnt == 0)
  {
    MyFree(arv);
    return;
  }

  cptr = GetCommand(vhostcmds, arv[0]);

  if (!cptr || (cptr == (struct Command *) -1))
  {
    notice(n_HostServ, lptr->nick,
      "%s command [%s]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      arv[0]);
    MyFree(arv);
    return;
  }

  /*
   * Check if the command is for admins only - if so,
   * check if they match an admin O: line.  If they do,
   * check if they are EITHER oper'd, or registered with OperServ,
   * if either of these is true, allow the command
   */
  if ((cptr->level == LVL_ADMIN) && !(IsValidAdmin(lptr)))
  {
    notice(n_HostServ, lptr->nick, "Unknown command [%s]",
      arv[0]);
    MyFree(arv);
    return;
  }

  /*
   * Use FindNick() here instead of GetLink() because
   * nicknames are identified individually, and using GetLink()
   * would tell us if the master of the link is identified,
   * not necessarily this specific nickname
   */
  if ((nptr = FindNick(lptr->nick)))
    if (nptr->flags & NS_FORBID)
    {
      notice(n_HostServ, lptr->nick,
        "Cannot execute commands for forbidden nicknames");
      MyFree(arv);
      return;
    }

  if (cptr->level != LVL_NONE)
  {
    if (!nptr)
    {
      notice(n_HostServ, lptr->nick,
        "You must register your nick with \002%s\002 first",
        n_NickServ);
      notice(n_HostServ, lptr->nick,
        "Type: \002/msg %s REGISTER <password>\002 to register your nickname", 
        n_NickServ);
      MyFree(arv);
      return;
    }

    if (!(nptr->flags & NS_IDENTIFIED))
    {
      notice(n_HostServ, lptr->nick,
        "Password identification is required for [\002%s\002]",
        cptr->cmd);
      notice(n_HostServ, lptr->nick, 
        "Type \002/msg %s IDENTIFY <password>\002 and retry",
        n_NickServ);
      MyFree(arv);
      return;
    }
  }

  /* call cptr->func to execute command */
  (*cptr->func)(lptr, acnt, arv);

  MyFree(arv);
} /* vs_process() */

/*
v_help()
 Give lptr help
*/

static void
v_help(struct Luser *lptr, int ac, char **av)

{
  if (ac >= 2)
  {
    char  str[MAXLINE];

    if (ac >= 3)
      ircsprintf(str, "%s %s", av[1], av[2]);
    else
    {
      if ((!irccmp(av[1], "ACCESS")) ||
          (!irccmp(av[1], "SHITLIST")))
        ircsprintf(str, "%s index", av[1]);
      else
      {
        struct Command *cptr;

        for (cptr = vhostcmds; cptr->cmd; cptr++)
          if (!irccmp(av[1], cptr->cmd))
            break;

        if (cptr->cmd)
          if ((cptr->level == LVL_ADMIN) &&
              !(IsValidAdmin(lptr)))
          {
            notice(n_HostServ, lptr->nick,
              "No help available on \002%s\002",
              av[1]);
            return;
          }

        ircsprintf(str, "%s", av[1]);
      }
    }

    GiveHelp(n_HostServ, lptr->nick, str, NODCC);
  }
  else
    GiveHelp(n_HostServ, lptr->nick, NULL, NODCC);
} /* v_help() */

/*
v_access()
  Modify the vhost list for lptr->nick
*/

static void
v_access(struct Luser *lptr, int ac, char **av)

{
  struct Command *cptr;
  struct NickInfo *target;

  if (ac < 2)
  {
    notice(n_HostServ, lptr->nick,
      "Syntax: \002ACCESS {ADD|DEL|LIST} [vhost]\002");
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "ACCESS");
    return;
  }

  target = NULL;

#ifdef EMPOWERADMINS_MORE

  /*
   * Allow administrators to specify a nickname to modify
   */
  if (IsValidServicesAdmin(lptr))
  {
    /*
     * First, check if av[1] is a valid command. If not,
     * check if it is a valid nickname. If av[1] is
     * a valid command, process the command as though
     * lptr is modifying their own access, which they
     * most likely are. If av[1] is neither a valid
     * command nor a valid nickname, give them
     * a "Nick not registered" error.
     */
    cptr = GetCommand(accesscmds, av[1]);
    if (!cptr || (cptr == (struct Command *) -1))
    {
      target = GetLink(av[1]);
      if (!target)
      {
        notice(n_HostServ, lptr->nick,
          ERR_NOT_REGGED,
          av[1]);
        return;
      }

      if (ac >= 3)
        cptr = GetCommand(accesscmds, av[2]);
      else
        cptr = NULL;
    }
  }
  else
    cptr = GetCommand(accesscmds, av[1]);

#else

  cptr = GetCommand(accesscmds, av[1]);

#endif /* EMPOWERADMINS_MORE */

  if (cptr && (cptr != (struct Command *) -1))
  {
    /* call the appropriate function */
    (*cptr->func)(lptr, target, ac, av);
  }
  else
  {
    /* the command doesn't exist */
    notice(n_HostServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[1]);
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "ACCESS");
  }
} /* v_access() */

/*
v_shit()
  Modify the shitlist
*/

static void
v_shit(struct Luser *lptr, int ac, char **av)

{
  struct Command *cptr;

  if (ac < 2)
  {
    notice(n_HostServ, lptr->nick,
      "Syntax: \002SHITLIST {ADD|DEL|LIST|SAVE} [regexp]\002");
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "SHITLIST");
    return;
  }

  cptr = GetCommand(shitcmds, av[1]);

  if (cptr && (cptr != (struct Command *) -1))
  {
    /* call the appropriate function */
    (*cptr->func)(lptr, ac, av);
  }
  else
  {
    /* the command doesn't exist */
    notice(n_HostServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[1]);
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "SHITLIST");
  }
} /* v_shit() */

static void
v_shit_add(struct Luser *lptr, int ac, char **av)

{
	char buf[MAXLINE];
	struct ShitList *tmpshit;

    if (ac < 3)
    {
      notice(n_HostServ, lptr->nick, "Syntax: \002SHITLIST ADD <regexp>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "SHITLIST ADD");
      return;
    }

    RecordCommand("%s: %s!%s@%s SHITLIST ADD %s",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[2]);

	tmpshit = (struct ShitList *) MyMalloc(sizeof(struct ShitList));
	strcpy(buf, av[2]);
	tmpshit->result = regcomp(&tmpshit->regex, buf, REG_ICASE | REG_NOSUB);
	
	if (tmpshit->result) {
		
		regerror(tmpshit->result, &tmpshit->regex, buf, MAXLINE);
		notice(n_HostServ, lptr->nick, 
			"regcomp() error on '%s' -> %s", av[2], buf);
			
		MyFree(tmpshit);
		return;
	} 

	tmpshit->pattern = MyStrdup(buf);
	tmpshit->author = MyStrdup(lptr->nick);
	tmpshit->next = shitlist;
	shitlist = tmpshit;

	notice(n_HostServ, lptr->nick, "Added '\002%s\002' to my shitlist", av[2]);
	
} /* v_shit_add() */

static void
v_shit_del(struct Luser *lptr, int ac, char **av)

{
	struct ShitList *tmpshit, *prev;

    if (ac < 3)
    {
      notice(n_HostServ, lptr->nick, "Syntax: \002SHITLIST DEL <regexp>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "SHITLIST DEL");
      return;
    }

    RecordCommand("%s: %s!%s@%s SHITLIST DEL %s",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[2]);

  	if (shitlist)
  	for (tmpshit = prev = shitlist; tmpshit; tmpshit = tmpshit->next) {

		if (!irccmp(tmpshit->pattern, av[2])) {

			prev->next = tmpshit->next;
			if (shitlist == tmpshit)
				shitlist = tmpshit->next;
			
			regfree(&tmpshit->regex);
			MyFree(tmpshit->pattern);
			MyFree(tmpshit->author);
			MyFree(tmpshit);
			
			notice(n_HostServ, lptr->nick, 
				"Deleted '\002%s\002' from my shitlist", av[2]);
			return;
		}
		prev = tmpshit;
		
	} else {

		notice(n_HostServ, lptr->nick, "No shitlist");
		return;
	}

	notice(n_HostServ, lptr->nick, "Couldn't find this in my shitlist");
	
} /* v_shit_del() */

static void
v_shit_list(struct Luser *lptr, int ac, char **av)

{
	int i = 0;
	struct ShitList *tmpshit;

    RecordCommand("%s: %s!%s@%s SHITLIST LIST",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

  	if (shitlist) {
  	
  		notice(n_HostServ, lptr->nick, "-- Shit List (regexp, author) --");
  		
  		for (tmpshit = shitlist; tmpshit; tmpshit = tmpshit->next)
			notice(n_HostServ, lptr->nick, "%3d) %-15s %-15s", 
				++i, 
				tmpshit->pattern,
				tmpshit->author? tmpshit->author : "");
				
		notice(n_HostServ, lptr->nick, "-- End of list --");

	} else {

		notice(n_HostServ, lptr->nick, "No shitlist");
	}
} /* v_shit_list() */

static void
v_shit_save(struct Luser *lptr, int ac, char **av)

{
	FILE *f;
	struct ShitList *tmpshit;

    RecordCommand("%s: %s!%s@%s SHITLIST SAVE",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

	if (shitlist) {
	
		if ((f = fopen(ShitFile? ShitFile : "shitlist", "w")) == NULL) {
		
			notice(n_HostServ, lptr->nick, "ERROR opening shitlist file '%s'",
				ShitFile? ShitFile : "shitlist");
			return;
		}
		
  		for (tmpshit = shitlist; tmpshit; tmpshit = tmpshit->next)
  			fprintf(f, "%s %s\n", 
  				tmpshit->pattern,
  				tmpshit->author? tmpshit->author : "");
  				
		fclose(f);
		
		notice(n_HostServ, lptr->nick, "Saved my shitlist");

	} else {

		notice(n_HostServ, lptr->nick, "No shitlist");
	}
} /* v_shit_save() */

void loadshitlist()
{
	char buf[MAXLINE];
	struct ShitList *tmpshit;
	FILE *f = fopen(ShitFile? ShitFile : "shitlist", "r");
	int ac; char **av;
	
	if (!f) return;
	
	if (shitlist)
	for (tmpshit = shitlist; tmpshit; ) {
		struct ShitList *sht = tmpshit;
		
		tmpshit = tmpshit->next;
		regfree(&sht->regex);
		MyFree(sht->pattern);
		MyFree(sht);
	}
	shitlist = NULL;
	
	while (fgets(buf, MAXLINE, f)) {
		char *p;
	
		if ((p = strchr(buf, '\n'))) *p=0;
		if ((p = strchr(buf, '\r'))) *p=0;
		
		if ((ac = SplitBuf(buf, &av)) == 0) {
			MyFree(av);
			continue;
		}
	
		tmpshit = (struct ShitList *) MyMalloc(sizeof(struct ShitList));
		tmpshit->next = shitlist;
		shitlist = tmpshit;
		
		tmpshit->result = regcomp(&tmpshit->regex, av[0], REG_ICASE | REG_NOSUB);
		tmpshit->pattern = MyStrdup(av[0]);
		tmpshit->author = ac>1? MyStrdup(av[1]) : NULL;
		
		if (tmpshit->result) {
		
			regerror(tmpshit->result, &tmpshit->regex, buf, MAXLINE);
			putlog(LOG1, "regcomp() error on '%s' -> %s", tmpshit->pattern, buf);
		}
		
		MyFree(av);
	}
	fclose(f);
}

static void
v_access_add(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  int num, i;
  struct NickInfo *nptr;
  struct NickVHost *vhptr;
  char *mask, *tmp;
  struct ShitList *tmpshit;

  if (target)
  {
    if (ac < 4)
    {
      notice(n_HostServ, lptr->nick,
        "Syntax: \002ACCESS <nickname> ADD <vhost>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "ACCESS ADD");
      return;
    }

    nptr = target;
    mask = av[3];
  }
  else
  {
    if (ac < 3)
    {
      notice(n_HostServ, lptr->nick, "Syntax: \002ACCESS ADD <vhost>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "ACCESS ADD");
      return;
    }

    nptr = GetLink(lptr->nick);
    mask = av[2];
  }

  if (strlen(mask) > HOSTLEN-1) {
	notice(n_HostServ, lptr->nick, "Vhost too long!");
        return;
  }

  if (!nptr)
    return;
    
  for (tmp = mask; *tmp; tmp++)
  	if (!isalnum(*tmp) && *tmp!='.' && *tmp!='-') {
        notice(n_HostServ, lptr->nick, "Invalid vhost!");
        notice(n_HostServ, lptr->nick, "Vhosts should only contain:");
        notice(n_HostServ, lptr->nick, "letters, digits, dots(.), and hyphens(-)");
        return;
  }
  
  if (!target) {
  
  	num = 0;
  	vhptr = nptr->vhosts;
  	if (VhostPerNick)
  	while (vhptr) {
  		if (++num == VhostPerNick) {
        	notice(n_HostServ, lptr->nick, "Your vhost list is full (%d vhosts)", num);
        	notice(n_HostServ, lptr->nick, "Vhost NOT added!");
			return;
		}  		
  		vhptr = vhptr->next;
  	}
  	
    num = 0;
  	vhptr = nptr->vhosts;
  	if (VhostChangePeriod)
  	while (vhptr) {
  		if (current_ts - vhptr->ts < VhostChangePeriod)
  			++num; 
  		vhptr = vhptr->next;
  	}
  	if (VhostChanges)
  	if (num >= VhostChanges) {
        notice(n_HostServ, lptr->nick, "Your vhost change quota exceeded");
        notice(n_HostServ, lptr->nick, "You are allowed max %d changes in %s",
        	VhostChanges, timeago(VhostChangePeriod, 3));
        notice(n_HostServ, lptr->nick, "Vhost NOT added!");
		return;
	}
  		
    i = 0;
    if (VhostChangePeriod)
    while (i < MAXVHCHANGES && nptr->vhchanges[i] 
    	&& (current_ts - nptr->vhchanges[i] < VhostChangePeriod))
    		++i;
    if (VhostChanges)
    if (i >= VhostChanges) {
      	notice(n_HostServ, lptr->nick, "Your vhost change limit exceeded");
      	notice(n_HostServ, lptr->nick, "You are allowed max %d changes in %s",
      		VhostChanges, timeago(VhostChangePeriod, 3));
      	notice(n_HostServ, lptr->nick, "Vhost NOT added!");
	  	return;
    }      	
    
  	for (num = 0; num < NICKLIST_MAX; num++) {
  	
  		struct NickInfo *nickp;
  		extern struct NickInfo *nicklist[];
  		
  		if (nicklist[num]) 
  		for (nickp = nicklist[num]; nickp; nickp = nickp->next) {
  		
  			vhptr = nickp->vhosts;
  			while (vhptr) {
  				if (!irccmp(vhptr->hostmask, mask)) {
  					notice(n_HostServ, lptr->nick, "Your proposed vhost is already being");
  					notice(n_HostServ, lptr->nick, "used by [\002%s\002], please choose "
  						"another", nickp->nick);
  					return;
  				}
  				vhptr = vhptr->next;	
  			}
  		}
  	}
  
  	if (shitlist)
  	for (tmpshit = shitlist; tmpshit; tmpshit = tmpshit->next) {
  	
  		if (!tmpshit->result)
  		if (regexec(&tmpshit->regex, mask, 0, NULL, 0) == 0) {
  		
  			notice(n_HostServ, lptr->nick, 
  				"Your proposed vhost matches a shitlisted pattern [\002%s\002]",
  				tmpshit->pattern);
  			notice(n_HostServ, lptr->nick,
  				"Please choose another, non-offensive & non-abusive vhost");
  			return;
  		}
  	}
  }
  
#ifdef HAVE_PTHREADS
  if (target) {
#endif
  		
  vhptr = (struct NickVHost *) MyMalloc(sizeof(struct NickVHost));
  vhptr->hostmask = MyStrdup(mask);
  vhptr->next = nptr->vhosts;
  nptr->vhosts = vhptr;
  
  vhptr->ts = current_ts;

  for (i=MAXVHCHANGES-1; i > 0; i--)
    nptr->vhchanges[i] = nptr->vhchanges[i-1];
  nptr->vhchanges[0] = current_ts;
  
#ifdef HAVE_PTHREADS
  }
#endif

  if (target)
  {
	notice(n_HostServ, lptr->nick,
       "[\002%s\002] has been added to the Vhost List for [%s]",
       mask, target->nick);

	RecordCommand("%s: Administrator %s!%s@%s ACCESS %s ADD [%s]",
	      n_HostServ,
	      lptr->nick,
	      lptr->username,
	      lptr->hostname,
	      target->nick,
	      mask);
  }
  else
  {
#ifdef HAVE_PTHREADS
	{
		pthread_t pthread;
		static void *args[5];
		
		args[0] = MyStrdup(lptr->nick); 
		args[1] = MyStrdup(nptr->nick); 
		args[2] = MyStrdup(mask); 
		args[3] = nptr; args[4] = lptr;
	
		notice(n_HostServ, lptr->nick,
	   		"Doing WHOIS (RFC954) lookup on your vhost, please wait...");
	   
		pthread_create(&pthread, NULL, (void *)whois_lookup, args);
	
		return;
	}
#else
	notice(n_HostServ, lptr->nick,
       "[\002%s\002] was added to your Vhost List",
       mask);

	RecordCommand("%s: %s!%s@%s ACCESS ADD [%s]",
	      n_HostServ,
	      lptr->nick,
	      lptr->username,
	      lptr->hostname,
	      mask);
#endif  
  }
  
} /* v_access_add() */

#ifdef HAVE_PTHREADS

static char *tokens[] = {
	"*not found*",
	"*no match*",
	"*no entries*",
	"*no data found*",
	"*not registered*",
	"*nothing found*",
	"*does not exist*",
	"*no such domain*",
	"*status*avail*",
	NULL
};

static char *notlds[] = {"gov", "cn", "nz", NULL};

static void 
whois_lookup(void *args)

{
	char buf[MAXLINE], *p;
	int num = 0, notreg = 0;
	char **pp;

	void **arr = args;
	char *lnick = (char *)arr[0];
	char *nnick = (char *)arr[1];
	char *vhost = (char *)arr[2];
	struct NickInfo *nptr = (struct NickInfo *)arr[3];
	struct Luser *lptr = (struct Luser *)arr[4];
	
	int oldstdout = dup(fileno(stdout));
	int oldstderr = dup(fileno(stderr));
	char *tmp = tmpnam(NULL);
	FILE *f = fopen(tmp, "w+");
	
	char *tld = strrchr(vhost, '.'); 
	
	if (tld) 
		tld++;
	else
		goto add;
		
	for (pp = notlds; *pp; pp++)
		if (!irccmp(*pp, tld)) {
		
			if (lptr == FindClient(lnick)) {
			
				notice(n_HostServ, lptr->nick,
					"You tried to add a vhost within forbidden TLD '%s'", *pp);
				notice(n_HostServ, lptr->nick,
					"Vhosts cannot use this TLD, please choose another");
			}
			goto bail;
		}
	
	for (p = tld; *p; p++)
		if (*p >= '0' && *p <= '9')
			goto add;

	p = tld; p--;
	while (p > vhost && *--p != '.');
	if (p > vhost)
		p++;
	
	if (tld && !irccmp(tld, "bg"))
		ircsprintf(buf, "whois -h whois.digsys.bg %s", p);
	else
		ircsprintf(buf, "whois %s", p);
	putlog(LOG1, "WHOIS lookup: %s", buf);
		
	dup2(fileno(f), fileno(stdout));
	dup2(fileno(f), fileno(stderr));
	system(buf);
	
	rewind(f);
	while (fgets(buf, MAXLINE, f)) {
		char **p = tokens;
	
		num++;
		while (*p) {
			if (match(*p, buf)) {
				notreg = 1;
				break;
			}
			p++;
		}
	}
	fclose(f); unlink(tmp);
	
	if (num == 1 || notreg) {
add:	
		if (nptr == FindNick(nnick)) {
			int i;
			struct NickVHost *vhptr;

  			vhptr = (struct NickVHost *) MyMalloc(sizeof(struct NickVHost));
  			vhptr->hostmask = MyStrdup(vhost);
  			vhptr->next = nptr->vhosts;
  			nptr->vhosts = vhptr;
  
  			vhptr->ts = current_ts;

  			for (i=MAXVHCHANGES-1; i > 0; i--)
    			nptr->vhchanges[i] = nptr->vhchanges[i-1];
  			nptr->vhchanges[0] = current_ts;
  			
  			if (lptr == FindClient(lnick)) {
  			
				notice(n_HostServ, lptr->nick,
       				"[\002%s\002] was added to your Vhost List",
       				vhost);

				RecordCommand("%s: %s!%s@%s ACCESS ADD [%s]",
	      			n_HostServ,
	      			lptr->nick,
	      			lptr->username,
	      			lptr->hostname,
	      			vhost);
  			}
		}
	
	} else if (num > 1) {
	
		if (lptr == FindClient(lnick)) {
			
			notice(n_HostServ, lptr->nick,
				"You tried to add a vhost which appears to be within existing domain");
			notice(n_HostServ, lptr->nick,
				"Vhosts can only feature an unregistered domains, please choose another");
		}
	}
bail:	
	dup2(oldstdout, fileno(stdout));
	dup2(oldstderr, fileno(stderr));
	close(oldstdout);
	close(oldstderr);
	MyFree(vhost);
	MyFree(nnick);
	MyFree(lnick);
	
} /* whois_lookup() */

#endif

static void
v_access_del(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  struct NickInfo *nptr;
  struct NickVHost *hptr, *prev;
  int found, i;
  char *mask;

  if (target)
  {
    if (ac < 4)
    {
      notice(n_HostServ, lptr->nick,
        "Syntax: \002ACCESS <nickname> DEL <vhost>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "ACCESS DEL");
      return;
    }

    nptr = target;
    mask = av[3];
  }
  else
  {
    if (ac < 3)
    {
      notice(n_HostServ, lptr->nick,
        "Syntax: \002ACCESS DEL <vhost>\002");
      notice(n_HostServ, lptr->nick,
        ERR_MORE_INFO, 
        n_HostServ,
        "ACCESS DEL");
      return;
    }

    nptr = GetLink(lptr->nick);
    mask = av[2];
  }

  if (!nptr)
    return;

  found = 0;

  prev = NULL;
  if (nptr->vhosts)
  for (hptr = nptr->vhosts; hptr; )
  {
    if (match(mask, hptr->hostmask))
    {
      found = 1;
      
      if (!target) {  
        i = 0;
        if (VhostChangePeriod)
        while (i < MAXVHCHANGES && nptr->vhchanges[i] 
      	  && (current_ts - nptr->vhchanges[i] < VhostChangePeriod))
      		  ++i;
        if (VhostChanges)
        if (i >= VhostChanges) {
          notice(n_HostServ, lptr->nick, "Your vhost change limit exceeded");
          notice(n_HostServ, lptr->nick, "You are allowed max %d changes in %s",
        	  VhostChanges, timeago(VhostChangePeriod, 3));
          notice(n_HostServ, lptr->nick, "Vhost NOT deleted!");
		  return;
        }
      }
      
      MyFree(hptr->hostmask);

      if (prev)
      {
        prev->next = hptr->next;
        MyFree(hptr);
        hptr = prev;
      }
      else
      {
        nptr->vhosts = hptr->next;
        MyFree(hptr);
        hptr = NULL;
      }
      
      for (i=MAXVHCHANGES-1; i > 0; i--)
      	nptr->vhchanges[i] = nptr->vhchanges[i-1];
      nptr->vhchanges[0] = current_ts;

      break;
    }

    prev = hptr;

    if (hptr)
      hptr = hptr->next;
    else
      hptr = nptr->vhosts;
  }

  if (found)
  {
    if (target)
    {
      notice(n_HostServ, lptr->nick,
        "[\002%s\002] removed from the Vhost List for [%s]",
        mask,
        target->nick);

      RecordCommand("%s: Administrator %s!%s@%s ACCESS %s DEL [%s]",
        n_HostServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        target->nick,
        mask);
    }
    else
    {
      notice(n_HostServ, lptr->nick,
        "[\002%s\002] removed from your Vhost List",
        mask);

      RecordCommand("%s: %s!%s@%s ACCESS DEL [%s]",
        n_HostServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        mask);
    }
  }
  else
  {
    if (target)
      notice(n_HostServ, lptr->nick,
        "[\002%s\002] was not found on the Vhost List for [%s]",
        mask,
        target->nick);
    else
      notice(n_HostServ, lptr->nick,
        "[\002%s\002] was not found on your Vhost List",
        mask);
  }
} /* v_access_del() */

static void
v_access_list(struct Luser *lptr, struct NickInfo *target, int ac, char **av)

{
  struct NickInfo *nptr;
  int cnt = 1;
  struct NickVHost *hptr;
  char *mask;

  if (target)
    nptr = target;
  else
    nptr = GetLink(lptr->nick);

  if (!nptr)
    return;

  mask = NULL;

  if (target)
  {
    if (ac >= 4)
      mask = av[3];
  }
  else
  {
    if (ac >= 3)
      mask = av[2];
  }

  if (target)
    RecordCommand("%s: Administrator %s!%s@%s ACCESS %s LIST %s",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      target->nick,
      mask ? mask : "");
  else
    RecordCommand("%s: %s!%s@%s ACCESS LIST %s",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      mask ? mask : "");

  notice(n_HostServ, lptr->nick,
    "-- VHost List for [\002%s\002] --",
    nptr->nick);

  for (hptr = nptr->vhosts; hptr; hptr = hptr->next)
  {
    if (mask)
      if (match(mask, hptr->hostmask) == 0)
        continue;

    notice(n_HostServ, lptr->nick, "%d) %s",
      cnt++,
      hptr->hostmask);
  }

  notice(n_HostServ, lptr->nick, "-- End of list --");
} /* v_access_list() */

/*
v_auth()
  Authorize GetLink(lptr->nick) to use SPOOF command
*/

static void
v_auth(struct Luser *lptr, int ac, char **av)

{
  struct NickInfo *nptr;
  char authcookie[20], buf[100];
  int i;

  if (ac < 2) {
  
  	nptr = GetLink(lptr->nick);
  	if (nptr->flags & NS_SPOOF) {
  	  notice(n_HostServ, lptr->nick, "You are already authorized to SPOOF");
  	  return;
  	}
  	  
  	if (nptr->authcookie) {
  	  notice(n_HostServ, lptr->nick, "Auth cookie has already been sent to you");
  	  notice(n_HostServ, lptr->nick, "Please check your e-mail for a message from me");
  	  return;
  	}
  	  
  	if (!nptr->email) {
  	  notice(n_HostServ, lptr->nick, "Please set your e-mail address first");
  	  notice(n_HostServ, lptr->nick, "Type \002/msg %s SET EMAIL <youremail>\002", n_NickServ);
  	  return;
  	}
  	
  	for (i=0; i < sizeof(authcookie)-1; i++)
  	  authcookie[i] = '0' + rand()%10;
  	authcookie[i] = 0;
  	
  	nptr->authcookie = MyStrdup(authcookie);
  	
  	ircsprintf(buf, "echo %s|mail -s \"Auth cookie\" %s -f%s@%s",
  		authcookie, nptr->email, n_HostServ, Me.name);
  	system(buf);
  	
  	notice(n_HostServ, lptr->nick, "I sent an auth cookie to your e-mail address");
  	notice(n_HostServ, lptr->nick, "Check your mailbox in a few minutes, and use");
  	notice(n_HostServ, lptr->nick, "the cookie to finish your authorization by typing:");
  	notice(n_HostServ, lptr->nick, "\002/msg %s AUTH <authcookie>\002", n_HostServ);

    RecordCommand("%s: %s!%s@%s AUTH cookie sent",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);
  	return;
  }
  
#ifdef EMPOWERADMINS_MORE

  if (ac > 2 && IsValidServicesAdmin(lptr)) {
  
  	nptr = GetLink(av[1]);
  	if (nptr) {
  	  nptr->flags |= NS_SPOOF;
  	  notice(n_HostServ, lptr->nick, "Authorized [\002%s\002] to use SPOOF", nptr->nick);
  	} else {
  	  notice(n_HostServ, lptr->nick, "[%s] is not a registered nick", av[1]);
  	  return;
  	}

    RecordCommand("%s: Administrator %s!%s@%s AUTH %s",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      nptr->nick);
  	return;
  	  
  } else if (ac > 2) {
  
    notice(n_HostServ, lptr->nick,
      "Syntax: \002AUTH <authcookie>\002");
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "AUTH");
    return;  
  }

#endif  

  nptr = GetLink(lptr->nick);
  if (nptr->flags & NS_SPOOF) {
  	notice(n_HostServ, lptr->nick, "You are authorized already to SPOOF");
  	return;
  }
  	  
  if (!nptr->authcookie) {
  	notice(n_HostServ, lptr->nick, "An auth cookie has still not been sent to you");
  	notice(n_HostServ, lptr->nick, "To make me email you one, type: \002/msg %s AUTH\002",
  		n_HostServ);
  	return;
  }
  	  
  if (!strcmp(nptr->authcookie, av[1])) {
  
  	nptr = GetLink(lptr->nick);
  	nptr->flags |= NS_SPOOF;
  	notice(n_HostServ, lptr->nick, "You have been authorized to use SPOOF, congratulations");
  	
    RecordCommand("%s: %s!%s@%s AUTH SUCCESS",
      n_HostServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

  } else {
  
  	notice(n_HostServ, lptr->nick, "The cookie you supplied doesn't match the one I sent to you");
  	notice(n_HostServ, lptr->nick, "Please check again your mailbox for a valid cookie");
  }
  
} /* v_auth() */

/*
v_spoof()
  Issue a SVSHOST command for lptr->nick
*/

static void
v_spoof(struct Luser *lptr, int ac, char **av)

{
  int found = 0;
  struct UsedVHost *usedptr = NULL;
  struct NickVHost *vhptr = NULL;
  char *p, buf[HOSTLEN + 1];

  if (ac < 2)
  {
    notice(n_HostServ, lptr->nick,
      "Syntax: \002SPOOF vhost\002");
    notice(n_HostServ, lptr->nick,
      ERR_MORE_INFO, 
      n_HostServ,
      "SPOOF");
    return;
  }
  
  if (GetLink(lptr->nick)->flags & NS_NOSPOOF) {
  	notice(n_HostServ, lptr->nick, "You are forbidden from using SPOOF, sorry");
  	return;
  }	
  
  if (!(GetLink(lptr->nick)->flags & NS_SPOOF)) {
  	notice(n_HostServ, lptr->nick, "You still have not used the AUTH command");
  	notice(n_HostServ, lptr->nick, "to get a proper authorization to use SPOOF");
  	notice(n_HostServ, lptr->nick, "Please type \002/msg %s HELP AUTH\002", n_HostServ);
  	return;
  }
  
  strcpy(buf, lptr->hostname);
  p = buf; while (*p++); --p;
  if (*p == '.') *p = 0;
  if (!irccmp(buf, av[1])) {
  	notice(n_HostServ, lptr->nick, "Won't spoof with your current hostname, doofus");
  	return;
  }
  
  if (!irccmp(av[1], "none"))
  	goto skip;
  
  if (GetLink(lptr->nick)->vhosts)
  for (vhptr = GetLink(lptr->nick)->vhosts; vhptr; vhptr = vhptr->next)
  	if (!irccmp(vhptr->hostmask, av[1]))
  	  found = 1;
  	  
  if (!found) {
    notice(n_HostServ, lptr->nick, "[\002%s\002] is NOT on your Vhost List", av[1]);
    return;
  }
  else
    found = 0;

  if (usedvh)
  for (usedptr = usedvh; usedptr; usedptr = usedptr->next)
  	if (!irccmp(usedptr->hostname, lptr->realhost)) {
  	
  	  int i = 0, same = 0;
      for (vhptr = usedptr->vhosts; vhptr; vhptr = vhptr->next, i++)
        if (!irccmp(vhptr->hostmask, av[1]))
          same = 1;
          
      if (!same) {
      
        if (VhostPerHost && i >= VhostPerHost) {
          notice(n_HostServ, lptr->nick, "Vhosts per host limit exceeded");
          notice(n_HostServ, lptr->nick, "Services allow max %d different vhosts", VhostPerHost);
          notice(n_HostServ, lptr->nick, "per host in %s", timeago(VhostPurgePeriod, 3));
          notice(n_HostServ, lptr->nick, "Please try again later");
          return;
        }
        
        vhptr = (struct NickVHost *) MyMalloc(sizeof(struct NickVHost));
        vhptr->hostmask = MyStrdup(av[1]);
        vhptr->next = usedptr->vhosts;
        usedptr->vhosts = vhptr;
        vhptr->ts = current_ts;
      }
      
      found = 1;
      break;
    }  	  
  
  if (!found) {
  
    vhptr = (struct NickVHost *) MyMalloc(sizeof(struct NickVHost));
    vhptr->hostmask = MyStrdup(av[1]);
    vhptr->ts = current_ts;
    vhptr->next = NULL;
    
    usedptr = (struct UsedVHost *) MyMalloc(sizeof(struct UsedVHost));
    usedptr->hostname = MyStrdup(lptr->realhost);
    usedptr->ts = current_ts;
    usedptr->vhosts = vhptr;
    
    if (usedvh) usedvh->prev = usedptr;
    usedptr->next = usedvh;
    usedptr->prev = NULL;
    usedvh = usedptr;
  }
  
  RecordCommand("%s: %s!%s@%s SPOOF %s",
		n_HostServ,
		lptr->nick,
		lptr->username,
		lptr->hostname,
		av[1]);

skip:
#ifdef BNET

	notice(n_HostServ, lptr->nick, "Sending SVSHOST to your server...");

	if (!irccmp(av[1], "none"))
	toserv(":%s SVSHOST %s %s :%s\n", 
		Me.name, 
		lptr->server->name,
		lptr->nick,
		lptr->realhost);
	else
	toserv(":%s SVSHOST %s %s :%s%s%s\n", 
		Me.name, 
		lptr->server->name,
		lptr->nick,
		VhostPrefix? VhostPrefix : "",
		av[1],
		VhostSuffix? VhostSuffix : "");

#endif 

} /* v_spoof() */

#endif /* HOSTSERVICES */
